// CatmullRom.cpp: implementation of the CCatmullRom class.
// Copyright: (c) Burkhard Wuensche
//////////////////////////////////////////////////////////////////////

#include "CatmullRom.h"
#include <gl/gl.h>
#include <gl/glu.h>
#include <gl/glut.h>


#define POINT_SIZE_CTRLPOINT 10.0
#define WIDTH_CURVE 3.0


double basis1(double t){ return -0.5*t*t*t+1.0*t*t-0.5*t;}
double basis2(double t){ return 1.5*t*t*t-2.5*t*t+1.0;}
double basis3(double t){ return -1.5*t*t*t+2.0*t*t+0.5*t;}
double basis4(double t){ return 0.5*t*t*t-0.5*t*t;}
double db1dt(double t){ return -1.5*t*t+2.0*t-0.5;}
double db2dt(double t){ return 4.5*t*t-5.0*t;}
double db3dt(double t){ return -4.5*t*t+4.0*t+0.5;}
double db4dt(double t){ return 1.5*t*t-t;}

typedef double (*basisFunctionType)(double);
basisFunctionType CRBasisFunction[]={&basis1,&basis2,&basis3,&basis4};
basisFunctionType CRBasisFunctionDerivative[]={&db1dt,&db2dt,&db3dt,&db4dt};

#define ORDER 4


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CCatmullRom::CCatmullRom()
{
	mat_ambient_and_diffuse[0]=0.5;
	mat_ambient_and_diffuse[1]=0.5;
	mat_ambient_and_diffuse[2]=0.3;
	mat_ambient_and_diffuse[3]=1;
	mat_specular[0]=0.6f;
	mat_specular[1]=0.6f;
	mat_specular[2]=0.6f;
	mat_specular[3]=1;
	mat_shininess[0]=100;
	glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mat_ambient_and_diffuse);
	glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);

}

CCatmullRom::~CCatmullRom()
{
}

void CCatmullRom::_computeCurvePoint(int segment, float t, CVec3df& p)
{	
	p.setVector(0.0, 0.0, 0.0);
	for(int i=0;i<ORDER;i++)
		p+=*ctrlPoints[segment+i]*CRBasisFunction[i](t);
}

void CCatmullRom::_computeCurveTangent(int segment, float t, CVec3df& v)
{	
	v.setVector(0.0, 0.0, 0.0);
	for(int i=0;i<ORDER;i++)
		v+=*ctrlPoints[segment+i]*CRBasisFunctionDerivative[i](t);
}


void CCatmullRom::computeCurveAndSurface(int& numPoints, CVec3df**& points)
{ 
	n = numPoints; 
	ctrlPoints = points; 

	int i,j;
	// compute NUM_CURVE_POINTS points on the Catmull-Rom spline defined by the "n" control points in the array "ctrlPoints"
	// Store the points in the array "curvePoints" defined in the constructor
	//
	// NOTE 1: The first and last control point are dummy points to ensure that the curve has the correct tangent at the start and end
	// That means the curve you draw should start at the second control point (curve segment defined by the points with index 0,1,2,3)
	// and it should end at the second to last control point (curve segment defined by the points with index numPoints-4, numPoints-3, numPoints-2, numPoints-1)
	// As a consequence you have n-3 segments, e.g. 4 control points define a curve with one segment, 5 control points define a curve with two segments etc.
	// 
	// NOTE 2: It is a good idea to also compute the derivative of the profile curve at the sample points
	// and store the values in the array "curveTangents"
	// The derivative is useful for computing the surface normal of the surface-of-revolution below 
	// You can use the formula in the lecture notes, but need to adjust it slightly to take into account that the rotation action is the y-axis and not the z-axis
	int numCurveSegments=n-3;
	float t;
	int segment;
	for(i=0;i<NUM_CURVE_POINTS;i++)
	{
		t=((float) numCurveSegments)*((float) i)/((float) (NUM_CURVE_POINTS-1));
		segment=(int) floor(t);
		t=t-segment;		
		if (segment<0) segment=0;
		if (segment>=numCurveSegments) segment=numCurveSegments-1;
		// the curve segment 'segment' has the control points with the indices segment, segment+1, segment+2, segment+3
		_computeCurvePoint(segment, t, curvePoints[i]);
		_computeCurveTangent(segment, t, curveTangents[i]);
	}

	// compute surface created by rotating Catmull-Rom spline around the x-axis
	// Store the vertices and vertex normals in the predefined arrays "surfacePoints" and "surfaceNormals"
	// defined in the constructor. Both arrays have a size of NUM_CURVE_POINTS x NUM_CIRCUMFERENTIAL_POINTS
	bool topToBottom = (curvePoints[0][Y] > curvePoints[NUM_CURVE_POINTS-1][Y]);

	double x,y,z,r,angle;
	CVec3df v1, v2, n;
	for(i=0;i<NUM_CURVE_POINTS;i++){
		cout << "\n curveTangents[i]=" << curveTangents[i];
		for(j=0;j<NUM_CIRCUMFERENTIAL_POINTS;j++){
			angle=2.0*Pi*j/((double) (NUM_CIRCUMFERENTIAL_POINTS-1));
			r=curvePoints[i][X];
			y=curvePoints[i][Y];
			x=r*cos(angle);
			z=r*sin(angle);
			surfacePoints[i][j].setVector(x,y,z);
			n.setVector(curveTangents[i][Y]*cos(angle), -curveTangents[i][X], curveTangents[i][Y]*sin(angle));
			n.normaliseDestructiveNoError();
			if (topToBottom) surfaceNormals[i][j]=n; else surfaceNormals[i][j]=-n;
		}
	}
}

void CCatmullRom::displayControlPoints()
{
	glPointSize(POINT_SIZE_CTRLPOINT);
	glColor3f(0,1,0);
	glBegin(GL_POINTS);
	for(int i=0;i<n;i++)
		glVertex3fv(ctrlPoints[i]->getArray());
	glEnd();
}

void CCatmullRom::displayCurve()
{
	// display the Catmull-Rom spline representing the profile curve using
	// a green line (RGB=(0,1,0)) with a width of WIDTH_CURVE 
	glLineWidth(WIDTH_CURVE);
	glColor3f(0,1,0);
	glBegin(GL_LINE_STRIP);
	for(int i=0;i<NUM_CURVE_POINTS;i++)
		glVertex3fv(curvePoints[i].getArray());
	glEnd();
}

void CCatmullRom::displaySurface()
{
	// display the surface using the materials properties defined below
	// Remember to use Gouraud shading (GL_SMOOTH) 
	// Please use the arrays "surfacePoints" and "surfaceNormals" and
	glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mat_ambient_and_diffuse);
	glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);

	glShadeModel(GL_SMOOTH);

	int i,j;
	for(i=0;i<(NUM_CURVE_POINTS-1);i++){
		glBegin(GL_QUAD_STRIP);
		for(j=0;j<NUM_CIRCUMFERENTIAL_POINTS;j++){
			glNormal3fv(surfaceNormals[i+1][j].getArray());
			glVertex3fv(surfacePoints[i+1][j].getArray());
			glNormal3fv(surfaceNormals[i][j].getArray());
			glVertex3fv(surfacePoints[i][j].getArray());
		}
		glEnd();
	}

}
